require 'OpenTK'
require 'System.Drawing'
include System::Drawing
include OpenTK
include OpenTK::Graphics
include OpenTK::Graphics::OpenGL


class MyGLWindowBase
  def initialize(width=640, height=480, title='window')
    @window = GameWindow.new(width, height, GraphicsMode.new(24, 24), title)
    @loaded = false
    @window.load.add( proc { eOtkLoad() } )
    @window.resize.add( proc { eOtkResize } )
    @window.update_frame.add( proc { eOtkUpdateFrame } )
    @window.render_frame.add( proc { eOtkPaint } )
    @window.Mouse.ButtonDown.add( proc { |sender,e| eOtkMouseButtonDown(sender, e) } )
    @window.Mouse.ButtonUp.add( proc { |sender,e| eOtkMouseButtonUp(sender, e) } )
    @window.Mouse.Move.add( proc { eOtkMouseMove } )
    @window.Mouse.WheelChanged.add( proc { eOtkMouseWheel } )
    #
    @eye_roll_x = 0
    @eye_roll_y = 0
    @eye_roll_z = 0
    @eye_distance = 3.0
    @eye_target_x = 0
    @eye_target_y = 0
    @eye_target_z = 0
    #
    @mouse_drag = false
    @mouse_pos_x = 0
    @mouse_pos_y = 0
    @mouse_wheel = @window.mouse.WheelPrecise
    #
    @refresh = true
  end

  def running
    @window.run(30.0, 0.0)
  end
  
  def camera(x, y, z)
    @eye_roll_x = x
    @eye_roll_y = y
    @eye_roll_z = z
  end

  def refresh
    @refresh = true
  end

  # --- display event

  def eOtkLoad
    GL.clear_color(Color.dark_green)
    GL.enable(EnableCap.depth_test)
    @loaded = true
    @refresh = true
  end
  
  def eOtkResize
    GL.viewport(0, 0, @window.width, @window.height)
    aspect_ratio = @window.width.to_f / @window.height

    perspective = OpenTK::Matrix4.create_perspective_field_of_view(
      MathHelper.PiOver4,
      aspect_ratio,
      1, 64)
    GL.matrix_mode(MatrixMode.projection)
    GL.load_matrix(perspective)
    @refresh = true
  end
  
  def eOtkPaint
    return if @loaded == false
    return if @refresh == false
    #
    @window.MakeCurrent()

    GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
    GL.MatrixMode(MatrixMode.Modelview);
    GL.load_matrix( cal_lookat() )

    draw()

    @window.swap_buffers
    @refresh = false
  end
  
  def eOtkUpdateFrame
    if @window.keyboard[OpenTK::Input::Key.escape] then
      @window.exit
    end

    tick()
  end
  
  # --- mouse event
  
  def eOtkMouseButtonDown(sender, e)
    @mouse_drag = true
    mouse_down(e)
  end
  
  def eOtkMouseButtonUp(sender, e)
    @mouse_drag = false
    mouse_up(e)
  end

  def eOtkMouseMove
    mouse_move()
  end

  def eOtkMouseWheel
    mouse_change_wheel()
  end

  # --- view point

  def deg2rad( deg )
    rad = Math::PI * 2 * deg / 360.0
  end

  def cal_lookat
    eye_x = 0
    eye_y = 0
    eye_z = @eye_distance
    rx = deg2rad(@eye_roll_x)
    ry = deg2rad(@eye_roll_y)
    rz = deg2rad(@eye_roll_z)
    # - roll x
    tx = eye_x
    ty = eye_y*Math.cos(rx) + eye_z*Math.sin(rx)
    tz = -eye_y*Math.sin(rx) + eye_z*Math.cos(rx)
    eye_x = tx
    eye_y = ty
    eye_z = tz
    # - roll y
    tx = eye_x*Math.cos(ry) - eye_z*Math.sin(ry)
    ty = eye_y
    tz = eye_x*Math.sin(ry) + eye_z*Math.cos(ry)
    eye_x = tx
    eye_y = ty
    eye_z = tz
    # - roll z
    #tx = eye_x*Math.cos(rz) + eye_y*Math.sin(rz)
    #ty = -eye_x*Math.sin(rz) + eye_y*Math.cos(rz)
    #tz = eye_z
    #eye_x = tx
    #eye_y = ty
    #eye_z = tz
    #
    frame_roll_x = Math.sin( rz )
    frame_roll_y = Math.cos( rz )
    lookat = Matrix4.look_at(eye_x, eye_y, eye_z,
                             @eye_target_x, @eye_target_y, @eye_target_z,
                             frame_roll_x,  frame_roll_y, 0)
  end


  # ----- override method
  
  # -----------------------------------------------------
  #  draw frame
  # -----------------------------------------------------

  def draw
  end  

  # -----------------------------------------------------
  #  tick work
  # -----------------------------------------------------

  def tick
  end  

  # -----------------------------------------------------
  #  mouse event
  # -----------------------------------------------------

  def mouse_down(e)
    @mouse_pos_x = @window.mouse.x
    @mouse_pos_y = @window.mouse.y
  end

  def mouse_up(e)
  end

  def mouse_move
    return if @mouse_drag == false
    mx = @window.mouse.x
    my = @window.mouse.y
    dx = (@mouse_pos_x - mx) / 2.0
    dy = (@mouse_pos_y - my) / 2.0
    @eye_roll_x = (@eye_roll_x - dy) % 360
    @eye_roll_y = (@eye_roll_y - dx) % 360
    @mouse_pos_x = mx
    @mouse_pos_y = my
    @refresh = true
  end

  def mouse_change_wheel
    move = @window.mouse.WheelPrecise
    wd = move - @mouse_wheel
    @eye_distance += (wd / 4)
    if @eye_distance < 0.1 then
      @eye_distance = 0.1
    end
    if @eye_distance > 1000.0 then
      @eye_distance = 1000.0
    end
    @mouse_wheel = move
    @refresh = true
  end

end

